(*

-----------------------------------------------------------------------------------------
                                     STATE
-----------------------------------------------------------------------------------------
 THIS SOFTWARE IS FREEWARE AND IS PROVIDED AS IS AND COMES WITH NO WARRANTY OF ANY
 KIND, EITHER EXPRESSED OR IMPLIED. IN NO EVENT WILL THE AUTHOR(S) BE LIABLE FOR
 ANY DAMAGES RESULTING FROM THE USE OF THIS SOFTWARE.
-----------------------------------------------------------------------------------------
                                  INFORMATION
-----------------------------------------------------------------------------------------
 Description : Contains the
               TbtMask component,
               TbtEdit component - a descendant of the TCustomEdit - and
               TbtComboEdit component - a descendant of RxLib's TCustomComboEdit

               -- TbtMask
               TbtMask is a kind of a controller component. It controls the text
               of a TCustomEdit descendant passed to its EditControl property.
               TbtMask forces the controlled EditControl to a behavior similar to
               VCL's TMaskEdit but with substantial differences in the way the text
               is masked. That behavior is more or less based on the Clipper's
               Get class. Check the comments in the interface section to see
               how a mask is contructed.
               There are three predefined masks (actually two), custom, numeric and date.

               -- TbtEdit
               it is just a plain TEdit with a few properties added such as
               Alignment, FontOnFocus and FontOnNotFocus

               -- TbtComboEdit
               it is just the plain TComboEdit of the great RxLib (thank you again guys)
               with a few properties added such as FontOnFocus, FontOnNotFocus and
               ButtonVisible.

 Tested      : Delphi 5, Win2K, RX Library 2.75
 More info   : Check the comments in the interface section
 License     : Freeware
 Thanks to   : RxLib creators
 Author      : Theo Bebekis, Thessaloniki, Greece <bebekis@otenet.gr>
-----------------------------------------------------------------------------------------
                                     HISTORY
-----------------------------------------------------------------------------------------
 Version   Date          Changes - Additions                                By
-----------------------------------------------------------------------------------------
 0.01      03.09.2000    Initial Version                                    

-----------------------------------------------------------------------------------------
   
*)

unit btMasks;


interface

{ un-comment the following line if you have RxLib installed }
//{$DEFINE RX_INSTALLED}

uses
  Windows
  ,Messages
  ,SysUtils
  ,Classes
  ,Graphics
  ,Controls
  ,StdCtrls
  {$IFDEF RX_INSTALLED}
  ,ToolEdit
  {$ENDIF}
  ;



const
  cFull     = [#0..#255];
  cBlanks   = [#0..#31];
  cAlpha    = [#32..#255]; { not all countries are using the standard ASCII chart }
  cNumbers  = ['0'..'9'];
  cBinary   = ['0','1'];
  cSigns    = ['+', '-'];
  cHex      = ['0'..'9'] + ['A'..'F'] + ['a'..'f'];

  (*
   X = any of the cAlpha is valid
   9 = any of the cNumbers is valid
   # = any of the cSigns or cNumbers is valid
   ! = any of the cAlpha is valid - coverts it to uppercase
  *)
  cMaskChars = ['X', '9', '#', '!'];



type
  TChars = set of char;
  TValidCharType = (ctSep, ctAlfa, ctDigit, ctDigitSign, ctUpper);
  TInsKeyState = (isInsert, isOverwrite);
  TMaskType = (mtCustom, mtNumeric, mtDate);
  TDirection = (drLeft, drRight);

  TRange = 1..100;           { 100 characters is enough for a mask }
  TSepPos = set of TRange;

  TOnExitValidationEvent = procedure (Sender: TObject; var Text: string) of object;


(*----------------------------------------------------------------------------------
 Description    : A field object stores info regarding the starting end ending char
                  of the field into a mask.
                  Given the mask
                    (XXX)-(XXX)
                  the MaskFields list of a TbtMask would contain 2 MaskField objects.
                  The fist MaskField object would be
                    First := 2;
                    Last  := 4;
                  where the second MaskField object would be
                    First := 8;
                    Last  := 10;
                  and the TbtMask.SepPos would be
                    SepPos := [1,5,6,7,11]   
-----------------------------------------------------------------------------------*)
  TMaskField = class
  public
    First    : integer;
    Last     : integer;
  end;
(*----------------------------------------------------------------------------------
 Description    : controls a TCustomEdit descendant applying a mask to its text
                  There are three predefined masks : custom, numeric and date
                  Assign properly the MaskType property for them to work.
                  Possible settings are
                    mtCustom,
                    mtNumeric,
                    mtDate

                  -- Numeric mask
                  The MaxLength of the controlled TCustomEdit is ignored.
                  Define the Size, Decimals and Signed properties, i.e the settings
                     Size     := 6;
                     Decimals := 2
                     Signed   := True;
                  construct the following mask
                     #99,99
                  which permits numbers as
                     -10.50
                     310.50
                  and so on

                  -- Date mask
                  The MaxLength of the controlled TCustomEdit is ignored.
                  The Size, Decimals and Signed properties are ignored
                  A Date mask has always the same followning shape
                  00/00/0000

                  -- Custom mask
                  The MaxLength of the controlled TCustomEdit is ignored.
                  You can construct any kind of mask you want, i.e
                  (XXX)-XXX-XXXX

                  --
                  If there is no mask at all, means the MaskText = ''
                  the controlled TCustomEdit doesn't change its normal behavior.

                  When the MaskType in [mtNumeric, mtDate] you can get/set the
                  Text of the controlled TCustomEdit as Variant using the
                  Value public property.
-----------------------------------------------------------------------------------*)
  TbtMask = class(TComponent)
  private
    FMasks         : TStrings;       { a TStringList where the class user might store additional masks }
    FMaskType      : TMaskType;
    MaskFields     : TStringList;    { holds a set of TMaskField objects }
    FMaskText      : string;         { the text of the mask i.e. (XXX)-(XXX) }
    FAutoWidth     : boolean;        { sets the width of the controlled TCustomEdit according to mask's size }
    SepPos         : TSepPos;        { a set of integers holding the position of the various separators in a mask i.e. SepPos := [1,5,6,7,11] }
    Format         : string;         { the text of the mask without the character placeholders i.e. (   )-(   ) }
    FSize          : Word;           { the size of a numeric mask }
    FDecimals      : Word;           { the decimals of a numeric mask }
    FSigned        : boolean;        { if True then the first char of a numeric mask could be + or - or cNumbers }
    DecSep         : Char;
    DateSep        : Char;
    OldText        : string;
    bFirstKey      : boolean;
    FClearOnEnter  : boolean;

    FEditControl   : TCustomEdit;
    hEdit          : THandle;
    pWndProcOld    : TFarProc;
    pWndProcNew    : TFarProc;

    FOnExitValidation : TOnExitValidationEvent;

    procedure SetEditControl(Value: TCustomEdit);
    procedure Edit_WndProc(var Message: TMessage);

    procedure SetNumericMask;
    function  CanAddChar(var C: Char; iPos: integer): boolean;
    procedure ClearFieldList;
    procedure SplitMaskText;
    function  FieldAtCaretPos(iPos: integer): TMaskField;
    procedure GetSeparators;
    procedure SetMasks(const V: TStrings);
    procedure SetMaskText(const V: string);
    procedure SetMaskType(const V: TMaskType);
    procedure SetDecimals(const V: Word);
    procedure SetSize(const V: Word);
    procedure SetSigned(const V: boolean);
    function  GetValue: Variant;
    procedure SetValue(const V: Variant);
    procedure CMWinIniChange(var Message: TMessage); message CM_WININICHANGE;
    procedure SetAutoWidth(const V: boolean);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure AddChar(iPos: integer; var Key: Char );
    procedure DelChar(iPos: integer; Direction: TDirection);
    procedure NumSepPressed(iPos: integer);
    procedure ValidateExit;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;          
    function  IsMasked: boolean;
    procedure Clear;
    property  Value           : Variant read GetValue write SetValue;      { the text of the controlled TCustomEdit as Variant }
  published
    property AutoWidth        : boolean read FAutoWidth write SetAutoWidth;
    property ClearOnEnter     : boolean read FClearOnEnter write FClearOnEnter;
    property EditControl      : TCustomEdit read FEditControl write SetEditControl;
    property MaskText         : string read FMaskText write SetMaskText;
    property Masks            : TStrings read FMasks write SetMasks;
    property MaskType         : TMaskType read FMaskType write SetMaskType;
    property Size             : Word read FSize write SetSize;
    property Decimals         : Word read FDecimals write SetDecimals;
    property Signed           : boolean read FSigned write SetSigned;
    property OnExitValidation : TOnExitValidationEvent read FOnExitValidation write FOnExitValidation;
  end;
(*----------------------------------------------------------------------------------
 Description    : it is just a plain TEdit with a few properties added such as
                  Alignment,
                  FontOnFocus and
                  FontOnNotFocus
                  I also have commented some other original properties because
                  I don't need them. You can un-comment them if you wish
-----------------------------------------------------------------------------------*)
  TbtEdit = class(TCustomEdit)
  private
    FAlignment      : TAlignment;
    FFontOnNotFocus : TFont;
    FFontOnFocus    : TFont;
    function GetCtl3D: boolean;
    procedure SetAlignment(const Value: TAlignment);
    procedure SetCtl3D(const Value: boolean);
    procedure SetFontOnFocus(const Value: TFont);
    procedure SetFontOnNotFocus(const Value: TFont);
  protected
    procedure CMEnter(var Message: TMessage); message CM_ENTER;
    procedure CMExit(var Message: TMessage); message CM_EXIT;
    procedure CreateParams(var Params: TCreateParams); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
  published
    property FontOnFocus      : TFont read FFontOnFocus write SetFontOnFocus    ;
    property FontOnNotFocus   : TFont read FFontOnNotFocus  write SetFontOnNotFocus ;
    property Alignment        : TAlignment read FAlignment write SetAlignment default taLeftJustify;
    
    property Anchors;
    property AutoSelect;
    property AutoSize;
    //property BiDiMode;
    property BorderStyle;
    property CharCase;
    property Color;
    property Constraints;
    property Control3D : boolean read GetCtl3D write SetCtl3D ;
    //property DragCursor;
    //property DragKind;
    //property DragMode;
    property Enabled;
    //property Font;
    property HideSelection;
    //property ImeMode;
    //property ImeName;
    property MaxLength;
    property OEMConvert;
    //property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    property OnChange;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    //property OnDragDrop;
    //property OnDragOver;
    //property OnEndDock;
    //property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    //property OnStartDock;
    //property OnStartDrag;
  end;
{$IFDEF RX_INSTALLED}
(*----------------------------------------------------------------------------------
 Description    : it is just the plain TComboEdit of the great RxLib (thank you again guys)
                  with a few properties added such as
                  FontOnFocus,
                  FontOnNotFocus and
                  ButtonVisible
                  I also have commented some other original properties because
                  I don't need them. You can un-comment them if you wish but I advise
                  you better check the code of the original TComboEdit first.
                  There was some conditional defines I've stripped here
-----------------------------------------------------------------------------------*)
  TbtComboEdit = class(TCustomComboEdit)
  private
    FButtonVisible: boolean;
    FFontOnNotFocus: TFont;
    FFontOnFocus: TFont;
    procedure SetButtonVisible(const Value: boolean);
    procedure SetFontOnFocus(const Value: TFont);
    procedure SetFontOnNotFocus(const Value: TFont);
  protected
    procedure CMEnter(var Message: TMessage); message CM_ENTER;
    procedure CMExit(var Message: TMessage); message CM_EXIT;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    property Button;
  published
    property FontOnFocus      : TFont read FFontOnFocus write SetFontOnFocus    ;
    property FontOnNotFocus   : TFont read FFontOnNotFocus  write SetFontOnNotFocus ;
    property ButtonVisible    : boolean read FButtonVisible write SetButtonVisible default False;

    property Alignment;
    property AutoSelect;
    property BorderStyle;
    property ButtonHint;
    property CharCase;
    property ClickKey;
    property Color;
    property Ctl3D;
    property DirectInput;
    property DragCursor;
    property DragMode;
    property EditMask;
    property Enabled;
    property Font;
    property GlyphKind;
    { Ensure GlyphKind is published before Glyph and ButtonWidth }
    property Glyph;
    property ButtonWidth;
    property HideSelection;
    property Anchors;
    //property BiDiMode;
    property Constraints;
    //property DragKind;
    //property ParentBiDiMode;
    //property ImeMode;
    //property ImeName;
    property MaxLength;
    property NumGlyphs;
    property OEMConvert;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    property OnButtonClick;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    //property OnDragOver;
    //property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    //property OnStartDrag;
    property OnContextPopup;
    //property OnEndDock;
    //property OnStartDock;
  end;
{$ENDIF}  // RX_INSTALLED


procedure Register;


implementation


{$R BTMASKS.RES}

uses
  Forms
  ;

  
(*----------------------------------------------------------------------------------*)
procedure Register;
begin
  RegisterComponents('xControls', [TbtMask]);
  RegisterComponents('xControls', [TbtEdit]);
{$IFDEF RX_INSTALLED}
  RegisterComponents('xControls', [TbtComboEdit]);
{$ENDIF}
end;


resourcestring
  SGetSetValueNotSupported = 'Value operations are supported ONLY with Numeric and Date masks';
  

type
  EbtMask = class(Exception);
{---------------------------------------------------------------------------------}
procedure RaiseError(const Msg: string);
begin
  raise EbtMask.Create(Msg);
end;
{---------------------------------------------------------------------------------}
procedure RaiseErrorFmt(const Msg: string; const Args: array of const);
begin
  raise EbtMask.CreateFmt(Msg, Args);
end;





{ helper routines }
(*----------------------------------------------------------------------------------*)
function GetInsKeyState: TInsKeyState;
begin
 if GetKeyState(VK_INSERT) = 0
  then Result := isInsert
  else Result := isOverWrite;
end;
(*----------------------------------------------------------------------------------*)
function Max(A, B: Integer): integer;
begin
  if A > B then Result := A else Result := B;
end;
{----------------------------------------------------------------------------------
 Description    : returns the current col (char) number of a windows edit control
                  in the line containing the caret
 Parameters     : a HWND of a windows edit control ( Memo, RichEdit, EditBox )
 Error checking : NO
 Notes          :
 Author         : Theo Bebekis <bebekis@otenet.gr>  <bebekis@x-treme.gr>
-----------------------------------------------------------------------------------}
function GetCurCol(Handle: HWND): integer;
begin
  Result := LoWord(SendMessage(Handle, EM_GETSEL, 0, 0)) - SendMessage(Handle, EM_LINEINDEX, -1, 0);
end;
{----------------------------------------------------------------------------------
 Description    : set the caret in the line specified by the Row parameter
                  and the character spedified by the Col parameter
                  of a windows edit control ( Memo, RichEdit )
 Error checking : NO
 Notes          :
 Author         : Theo Bebekis <bebekis@otenet.gr>  <bebekis@x-treme.gr>
-----------------------------------------------------------------------------------}
procedure SetCaretPos(Handle: HWND; Col, Row: Longword);
var
 CharIndex : integer;
begin
  CharIndex := SendMessage(Handle, EM_LINEINDEX, Row, 0)  + Integer(Col);
  SendMessage(Handle, EM_SETSEL, CharIndex, CharIndex);
end;
(*----------------------------------------------------------------------------------*)
function GetTextWidth(Font: TFont): Integer;
var
  DC: HDC;
  SaveFont: HFont;
  SysMetrics, Metrics: TTextMetric;
begin
  DC := GetDC(0);
  try
    GetTextMetrics(DC, SysMetrics);
    SaveFont := SelectObject(DC, Font.Handle);
    GetTextMetrics(DC, Metrics);
    SelectObject(DC, SaveFont);
  finally
    ReleaseDC(0, DC);
  end;
  Result := Max(SysMetrics.tmAveCharWidth, Metrics.tmAveCharWidth);
end;
(*----------------------------------------------------------------------------------*)
function GetValidCharType(C : Char): TValidCharType;
begin
  Result := ctSep;
  case C of
    'X' : Result := ctAlfa       ;
    '9' : Result := ctDigit      ;
    '#' : Result := ctDigitSign  ;
    '!' : Result := ctUpper      ;
  end;
end;





type
  THackEdit = class(TCustomEdit); { for accessing protected properties of the controlled TCustomEdit }


 { TbtMask }
(*----------------------------------------------------------------------------------*)
constructor TbtMask.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  hEdit := 0;
  FEditControl := nil;
  pWndProcNew  := nil;
  OldText      := '';
  bFirstKey    := True;
  FMaskType    := mtCustom;
  FMasks       := TStringList.Create;
  with FMasks do
  begin
    Add('(9999)-(999999)');
    Add('(999)-(999999)');
  end;
  MaskFields   := TStringList.Create;
  FSize        := 11;
  FDecimals    := 0;
  FSigned      := True;
  GetSeparators;
end;
(*----------------------------------------------------------------------------------*)
destructor TbtMask.Destroy;
begin  
  ClearFieldList;
  MaskFields.Free;
  Masks.Free;
  inherited Destroy;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.ClearFieldList;
begin
  Format := '';
  while MaskFields.Count > 0 do
  begin
    TMaskField(MaskFields.Objects[0]).Free;
    MaskFields.Delete(0);
  end;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.Notification(AComponent: TComponent;  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FEditControl) then
    FEditControl := nil;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.Clear;
begin
  if EditControl = nil then Exit;
  EditControl.Clear;
  if IsMasked then
    EditControl.Text := Format
  else  EditControl.Text := '';
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.SetEditControl(Value: TCustomEdit);
begin
  FEditControl := Value;
  if Value <> nil then
  begin
    Value.FreeNotification(Self);
    pWndProcNew := MakeObjectInstance(Edit_WndProc);
    hEdit       := TWinControl(Value).Handle;
    pWndProcOld := Pointer(GetWindowLong(hEdit, GWL_WNDPROC));
    SetWindowLong(hEdit, GWL_WNDPROC, LongInt(pWndProcNew));
    if IsMasked then
    begin
      SplitMaskText;
      SetAutoWidth(AutoWidth); 
    end;
  end
  else
  begin
    SetWindowLong(hEdit, GWL_WNDPROC, LongInt(pWndProcOld));
    hEdit := 0;
    pWndProcOld  := nil;
    FreeObjectInstance(pWndProcNew);
  end;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.SetAutoWidth(const V: boolean);
var
  iSize : integer;
begin
  FAutoWidth := V;
  if not FAutoWidth then Exit;
  if not IsMasked then Exit;
  if EditControl = nil then Exit;
  iSize := Length(Format);
  EditControl.Width := (iSize + 1) * GetTextWidth(THackEdit(EditControl).Font);
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.SetMasks(const V: TStrings);
begin
  FMasks.Assign(V);
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.SetDecimals(const V: Word);
begin
  FDecimals := V;
  if MaskType <> mtNumeric then Exit;
  SetNumericMask;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.SetSize(const V: Word);
begin
  FSize := V;
  if FSize = 0 then
    FSize := 11;
  if MaskType <> mtNumeric then Exit;
  SetAutoWidth(AutoWidth);
  SetNumericMask;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.SetSigned(const V: boolean);
begin
  FSigned := V;
  if MaskType <> mtNumeric then Exit;
  SetNumericMask;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.CMWinIniChange(var Message: TMessage);
begin
  inherited;
  GetSeparators;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.GetSeparators;
begin
  DecSep   := DecimalSeparator;
  DateSep  := DateSeparator;
end;
(*----------------------------------------------------------------------------------*)
function TbtMask.IsMasked: boolean;
begin
  Result := (MaskText <> '');
end;
(*----------------------------------------------------------------------------------*)
function  TbtMask.GetValue: Variant;
begin
  Result := Unassigned;
  if not IsMasked then Exit;
  if OldText <> '' then
    case MaskType of
      mtCustom   : RaiseError(SGetSetValueNotSupported);
      mtNumeric  : try
                     if Decimals > 0 then
                       Result := StrToInt(OldText)
                     else Result := StrToFloat(OldText);
                   except
                     SysUtils.Abort;
                   end;
      mtDate     : try
                     Result := StrToDate(OldText);
                   except
                     SysUtils.Abort;
                   end;
    end;
end;
(*----------------------------------------------------------------------------------*)
procedure   TbtMask.SetValue(const V: Variant);
var
  I : int64;
  F : Extended;
  D : TDate;
  S : string;
  sIntPart : string;
  sDecPart : string;
  iSep  : integer;
  iLen  : integer;
begin
  if not IsMasked then Exit;
  if EditControl = nil then Exit;

  case MaskType of
    mtCustom   : RaiseError(SGetSetValueNotSupported);
    mtNumeric  : begin
                   if Decimals = 0 then
                   begin
                     I := Integer(V);
                     S := IntToStr(I);
                     if Length(S) > Size then
                       SetLength(S, Size)
                     else S := StringOfChar(' ', Size - Length(S)) + S;
                   end
                   else
                   begin
                     F := V;
                     S := FloatToStrF(F, ffFixed, 18, Decimals);

                     if Length(S) > Size then
                     begin
                       iSep     := Pos(DecSep, S);
                       sDecPart := Copy(S, iSep, Length(S) - iSep + 1);   
                       sIntPart := IntToStr(Trunc(F));
                       iLen     := Size - Length(sDecPart);

                       if Length(sIntPart) > iLen then
                         SetLength(sIntPart, iLen);

                       S := sIntPart + sDecPart;
                       S := StringOfChar(' ', Size - Length(S)) + S;
                     end
                     else S := StringOfChar(' ', Size - Length(S)) + S;
                   end;
                 end;
    mtDate     : begin
                   D := VarToDateTime(V);
                   S := FormatDateTime('dd' + DateSep + 'mm' + DateSep + 'yyyy', D);
                 end;
  end;
  EditControl.Text := S;
  OldText := S;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.SetNumericMask;
var
  S : string;
begin
  S := '';
  try
    if Size = 0 then RaiseError('Zero size is invalid');
    if not (FDecimals in [0..9]) then RaiseError('Valid range for Decimals is 0 to 9');   
    if ((FMaskType = mtNumeric) and (FDecimals > 0)) then
      if FDecimals >= (FSize - 1) then RaiseError('Decimals >= Size - 1');

    S := StringOfChar('9', Size);
    if FDecimals > 0 then S[FSize - FDecimals] := DecSep;      
    if FSigned then S[1] := '#';
  finally
    FMaskText := S;
    SplitMaskText;
  end;
end;     
(*----------------------------------------------------------------------------------*)
procedure TbtMask.SetMaskType(const V: TMaskType);
begin   
  if FMaskType = V then Exit;
  FMaskType := V;
  case FMaskType of
    mtCustom  : FMaskText := '';
    mtNumeric : SetNumericMask;
    mtDate    : FMaskText := '99' + DateSep + '99' + DateSep + '9999';
  end;
  if (FMaskType in [mtNumeric, mtDate]) then
    ClearOnEnter := True;
  SplitMaskText;
  SetAutoWidth(AutoWidth);
  OldText := Format;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.SetMaskText(const V: string);
begin
  if FMaskType <> mtCustom then Exit;
  if V = FMaskText then Exit;
  FMaskText := V;
  SplitMaskText;
  SetAutoWidth(AutoWidth);
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.SplitMaskText;

   {----------------------------------------------------------------}
   procedure AddField(var S: string; var First: integer; i: integer);
   var
     MaskField : TMaskField;
   begin
     MaskField          := TMaskField.Create;
     MaskField.First    := First + 1;
     MaskField.Last     := i;
     MaskFields.AddObject('', MaskField);
     First := i;
     S := '';
   end;
   {----------------------------------------------------------------}

var 
  i         : integer;
  Len       : integer;
  S         : string;
  First     : integer;
begin

  ClearFieldList;
  if not IsMasked then Exit;
  Len := Length(FMaskText);

  if Len = 0 then
  begin
    if EditControl <> nil then
      EditControl.Text := '';
    Exit;
  end;

  SetLength(Format, Len);

  First   := 0;
  S       := '';
  SepPos  := [];

  for i := 1 to Len do
  begin
    if not (FMaskText[i] in cMaskChars) then
    begin
      Format[i] := FMaskText[i];
      Include(SepPos, i);     
      if S = '' then
      begin
        First := i;
        Continue;
      end;     
      AddField(S, First, i);
    end
    else
    begin
      S := S + FMaskText[i];
      Format[i] :=  ' ';  
    end;
  end;

  if S <> '' then AddField(S, First, i);
  if EditControl <> nil then
    EditControl.Text := Format;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.NumSepPressed(iPos: integer);
var
  S    : string;
  iSep : integer;
  Temp : string;
begin
  if EditControl = nil then Exit;
  if FDecimals = 0 then Exit;

  iSep := Pos(DecSep, FMaskText);

  S := EditControl.Text;

  if (iPos > iSep) then Exit;
  if (S[iSep - 1] <> ' ') then Exit;

  Temp   := Trim(Copy(S, 1, iSep - 1));
  S      := Temp + Copy(S, iSep, Size - (iSep - 1));
  S      := StringOfChar(' ', Size - Length(S)) + S;

  EditControl.Text := S;
  if (EditControl is TEdit) then
    SetCaretPos(EditControl.Handle, iSep, 1)
  else SetCaretPos(EditControl.Handle, iSep + 1, 1)

end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.ValidateExit;
var
  Temp      : string;
  S         : string;
  sIntPart  : string;
  sDecPart  : string;
  i         : integer;
  iSep      : integer;
  sDateF    : string;
begin
  if EditControl = nil then Exit;
  Temp := EditControl.Text;
  if Temp = Format then Exit;
  
  case FMaskType of
    mtCustom  : ;
    mtNumeric : begin
                  { decimal part }
                  iSep := Pos(DecSep, FMaskText);
                  if iSep <> 0 then
                  begin
                    sDecPart := Copy(Temp, iSep, Length(Temp) - iSep + 1);
                    for i := 1 to Length(sDecPart) do
                      if sDecPart[i] = ' ' then
                        sDecPart[i] := '0';
                  end;        

                  { integer part }
                  Temp := TrimLeft(Temp);
                  for i := 1 to Length(Temp) do
                  begin
                    if not (Temp[i] in [' ', DecSep]) then
                      sIntPart := sIntPart + Temp[i]
                    else //Break;
                    begin
                      if Decimals > 0 then
                      begin
                        SetLength(sDecPart, Decimals + 1);
                        sDecPart := DecSep + StringOfChar('0', Decimals);
                      end;
                      Break;
                    end;
                  end;
                  if sIntPart = '' then sIntPart := '0';                   

                  S := sIntPart + sDecPart;
                  S := StringOfChar(' ', Length(EditControl.Text) - Length(S)) + S;

                end;
    mtDate    : begin
                  sDateF := 'dd' + DateSep + 'mm' + DateSep + 'yyyy';
                  if Temp <> Format then
                  begin
                    for i := 1 to Length(Temp) do
                      if Temp[i] <> ' ' then
                      S := S + Temp[i];
                    S := FormatDateTime(sDateF, StrToDate(S));
                  end
                  else S := FormatDateTime(sDateF, StrToDate(Temp));
                end;
  end;
  if Assigned(FOnExitValidation) then FOnExitValidation(Self, S);
  EditControl.Text := S;
end;
(*----------------------------------------------------------------------------------*)
function  TbtMask.FieldAtCaretPos(iPos: integer): TMaskField;
var
  i         : integer;
  MaskField : TMaskField;
begin
  Result := nil;
  for i := 0 to MaskFields.Count - 1 do
  begin
    MaskField := TMaskField(MaskFields.Objects[i]);
    if (iPos in [MaskField.First..MaskField.Last]) then
    begin
      Result := MaskField;
      Exit;
    end;
  end;
end;
(*----------------------------------------------------------------------------------*)
function TbtMask.CanAddChar(var C: Char; iPos: integer): boolean;
var
  ValidCharType  : TValidCharType;
begin
  Result := False;
  ValidCharType := GetValidCharType(MaskText[iPos]);
  if ValidCharType = ctSep then Exit;
  case ValidCharType of
    ctAlfa       : if (C in cAlpha)            then Result := True;
    ctDigit      : if (C in cNumbers)          then Result := True;
    ctDigitSign  : if (C in cNumbers + cSigns) then Result := True;
    ctUpper      : if (C in cAlpha)            then Result := True;
  end;
  if (Result and (ValidCharType = ctUpper)) then C := UpCase(C);
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.AddChar(iPos: integer; var Key: Char );
var
  MaskField      : TMaskField;
  C              : Char;
  S              : string;
  i              : integer;
begin
  if EditControl = nil then Exit;
  C         := Key;
  Key       := #0;

  iPos := iPos + 1;

  MaskField := nil;
  while True do
  begin
    MaskField := FieldAtCaretPos(iPos);
    if (MaskField <> nil) then
      if CanAddChar(C, iPos) then
        Break;
    Inc(iPos);
    if iPos > Length(MaskText) then
      Exit;
  end;
          
  SetCaretPos(EditControl.Handle, iPos, 1);

  S := EditControl.Text;
  case getInsKeyState of
    isInsert    : begin
                    for i := MaskField.Last -1 downto iPos + 1 do
                      S[i] := S[i-1];
                    S[iPos] := C;
                  end;
    isOverWrite : begin
                    S := EditControl.Text;
                    S[iPos] := C;
                  end;
  end;

  EditControl.Text := S;
  if (EditControl is TEdit) then
    SetCaretPos(EditControl.Handle, iPos, 1)
  else SetCaretPos(EditControl.Handle, iPos + 1, 1)
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.DelChar(iPos: integer; Direction: TDirection);
var
  MaskField      : TMaskField;
  S              : string;
  i              : integer;
begin
  if EditControl = nil then Exit;
  S := EditControl.Text;

  case Direction of
    drLeft  : begin
                while (iPos in SepPos) do
                begin
                  Dec(iPos);
                  if iPos <= 0 then Exit;
                end;
              end;
    drRight : begin
                Inc(iPos);
                while (iPos in SepPos) do
                begin
                  Inc(iPos);
                  if iPos > Length(MaskText) then Exit;
                end;
              end;
  end;

  MaskField := FieldAtCaretPos(iPos);
  if MaskField = nil then Exit;

  for i := iPos to MaskField.Last - 1 do
  begin
    S[i] := S[i+1];
  end;
  
  S[MaskField.Last -1] := ' ';
  EditControl.Text := S;

  if (EditControl is TEdit) then
    SetCaretPos(EditControl.Handle, iPos - 1, 1)
  else SetCaretPos(EditControl.Handle, iPos, 1)
  
end;
(*----------------------------------------------------------------------------------*)
procedure TbtMask.Edit_WndProc(var Message: TMessage);
var
 iPos         : integer;
 C            : Char;
 Key          : Word;
begin
  if EditControl = nil then Exit;
  case Message.Msg of
     WM_KEYDOWN : begin
                    if not IsMasked then
                    begin
                      Message.Result := CallWindowProc(pWndProcOld, hEdit,  Message.Msg,  Message.WParam,  Message.LParam);
                      Exit;
                    end;

                    Key := TWMKey(Message).CharCode;

                    if (FClearOnEnter and bFirstKey and (Key <> VK_ESCAPE)) then
                    begin
                      Clear;
                      bFirstKey := False;
                    end;

                    case Key of
                       VK_ESCAPE : begin
                                     Message.Result := 0;
                                     EditControl.Text := OldText;
                                     Exit;
                                   end;
                       VK_DELETE : begin
                                     Message.Result := 0;
                                     iPos := GetCurCol(EditControl.Handle);
                                     DelChar(iPos, drRight);
                                   end;
                       else Message.Result := CallWindowProc(pWndProcOld, hEdit,  Message.Msg,  Message.WParam,  Message.LParam);
                    end;

                  end;
     WM_CHAR    : begin
                    if not IsMasked then
                    begin
                      Message.Result := CallWindowProc(pWndProcOld, hEdit,  Message.Msg,  Message.WParam,  Message.LParam);
                      Exit;
                    end;

                    Message.Result := 0;
                    Key  := TWMKey(Message).CharCode;
                    C    := Char(Key);
                    iPos := GetCurCol(EditControl.Handle);

                    if  (((C in ['.', ',']) or (C = DecimalSeparator)) and  (MaskType = mtNumeric)) then
                    begin
                      NumSepPressed(iPos);
                      Exit;
                    end;

                    if Ord(C) = VK_BACK then
                      DelChar(iPos, drLeft)
                    else AddChar(iPos, C);
                  end;    
     CM_ENTER   : begin
                    Message.Result := CallWindowProc(pWndProcOld, hEdit,  Message.Msg,  Message.WParam,  Message.LParam);
                    bFirstKey := True;
                    EditControl.SelLength := 0;
                  end;
     CM_EXIT    : begin
                    if IsMasked then
                      ValidateExit;
                    OldText := EditControl.Text;
                    Message.Result := CallWindowProc(pWndProcOld, hEdit,  Message.Msg,  Message.WParam,  Message.LParam);
                  end;
     else       Message.Result := CallWindowProc(pWndProcOld, hEdit,  Message.Msg,  Message.WParam,  Message.LParam);
  end;

end;







{ TbtEdit }
(*----------------------------------------------------------------------------------*)
constructor TbtEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);  
  FFontOnFocus       := TFont.Create;
  FFontOnNotFocus    := TFont.Create;
  FFontOnFocus.Style := [fsBold];
  Alignment          := taLeftJustify;
end;
(*----------------------------------------------------------------------------------*)
destructor TbtEdit.Destroy;
begin
  FFontOnFocus     .Free;
  FFontOnNotFocus  .Free;
  inherited;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtEdit.CreateParams(var Params: TCreateParams);
const
  Alignments: array[TAlignment] of Longword = (ES_LEFT, ES_RIGHT, ES_CENTER);
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style or ES_MULTILINE or WS_CLIPCHILDREN  or Alignments[FAlignment];
end;
(*----------------------------------------------------------------------------------*)
procedure TbtEdit.SetFontOnFocus(const Value: TFont);
begin
  FFontOnFocus.Assign(Value);
end;
(*----------------------------------------------------------------------------------*)
procedure TbtEdit.SetFontOnNotFocus(const Value: TFont);
begin
  FFontOnNotFocus.Assign(Value);
end;
(*----------------------------------------------------------------------------------*)
procedure TbtEdit.CMEnter(var Message: TMessage);
begin
  inherited;
  Font.Assign(FFontOnFocus);
end;
(*----------------------------------------------------------------------------------*)
procedure TbtEdit.CMExit(var Message: TMessage);
begin
  inherited;
  Font.Assign(FFontOnNotFocus);
end;
(*----------------------------------------------------------------------------------*)
function TbtEdit.GetCtl3D: boolean;
begin
  Result := inherited Ctl3D;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtEdit.SetCtl3D(const Value: boolean);
begin
  inherited Ctl3D := Value;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtEdit.SetAlignment(const Value: TAlignment);
begin
  if FAlignment <> Value then
  begin
    FAlignment := Value;
    RecreateWnd;
  end;
end;




{ TbtComboEdit }
{$IFDEF RX_INSTALLED}
(*----------------------------------------------------------------------------------*)
constructor TbtComboEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FFontOnFocus       := TFont.Create;
  FFontOnNotFocus    := TFont.Create;
  FFontOnFocus.Style := [fsBold];
  FButtonVisible     := False;
  Alignment          := taLeftJustify;    
  ControlState := ControlState + [csCreating];
  try
    ButtonWidth := 0;
  finally
    ControlState := ControlState - [csCreating];
  end;
end;
(*----------------------------------------------------------------------------------*)
destructor TbtComboEdit.Destroy;
begin
  FFontOnFocus     .Free;
  FFontOnNotFocus  .Free;
  inherited;
end;
(*----------------------------------------------------------------------------------*)
procedure TbtComboEdit.CMEnter(var Message: TMessage);
begin
  inherited;
  Font.Assign(FFontOnFocus);
end;
(*----------------------------------------------------------------------------------*)
procedure TbtComboEdit.CMExit(var Message: TMessage);
begin
  inherited;
  Font.Assign(FFontOnNotFocus);
end;
(*----------------------------------------------------------------------------------*)
procedure TbtComboEdit.SetButtonVisible(const Value: boolean);
begin
  if FButtonVisible = Value then Exit;
  FButtonVisible := Value;
  case FButtonVisible of
    True  : ButtonWidth := Max(Button.Width, 21);
    False : ButtonWidth := 0;
  end
end;
(*----------------------------------------------------------------------------------*)
procedure TbtComboEdit.SetFontOnFocus(const Value: TFont);
begin
  FFontOnFocus.Assign(Value);
end;
(*----------------------------------------------------------------------------------*)
procedure TbtComboEdit.SetFontOnNotFocus(const Value: TFont);
begin
  FFontOnNotFocus.Assign(Value);
end;
{$ENDIF}  // RX_INSTALLED
    
end.
